/* PeerCheckTasks - TimerTask that checks for good/bad up/downloaders.
   Copyright (C) 2003 Mark J. Wielaard

   This file is part of Snark.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.klomp.snark;

import java.util.*;

/**
 * TimerTask that cecks for good/bad up/downloader. Works together
 * with the PeerCoordinator to select which Peers get (un)choked.
 */
class PeerCheckerTask extends TimerTask
{
  private static final boolean VERBOSE = false;

  private final long KILOPERSECOND = 1024*(PeerCoordinator.CHECK_PERIOD/1000);

  private final PeerCoordinator coordinator;

  PeerCheckerTask(PeerCoordinator coordinator)
  {
    this.coordinator = coordinator;
  }

  public void run()
  {
    synchronized(coordinator.peers)
      {
	// Calculate total uploading and worst downloader.
	long worstdownload = Long.MAX_VALUE;
	Peer worstDownloader = null;

	int uploaders = 0;
	int downloaders = 0;
	int interested = 0;
	int interesting = 0;
	int choking = 0;
	int choked = 0;

	long uploaded = 0;
	long downloaded = 0;

	// Keep track of peers we remove now,
	// we will add them back to the end of the list.
	List removed = new ArrayList();

	Iterator it = coordinator.peers.iterator();
	while (it.hasNext())
	  {
	    Peer peer = (Peer)it.next();

	    if (peer.isUpLoading())
	      uploaders++;
	    if (peer.isDownLoading())
	      downloaders++;
	    if (peer.isInterested())
	      interested++;
	    if (peer.isInteresting())
	      interesting++;
	    if (peer.isChoking())
	      choking++;
	    if (peer.isChoked())
	      choked++;

	    long upload = peer.getUploaded();
	    uploaded += upload;
	    long download = peer.getDownloaded();
	    downloaded += download;

	    if (VERBOSE)
	      {
		System.out.println(peer + ":");
		System.out.println(" ul: " + upload/KILOPERSECOND
				   + " dl: " + download/KILOPERSECOND
				   + " i: " + peer.isInterested()
				   + " I: " + peer.isInteresting()
				   + " c: " + peer.isChoking()
				   + " C: " + peer.isChoked()
				   + " u: " + peer.isUpLoading()
				   + " d: " + peer.isDownLoading());
	      }

	    peer.resetCounters();

	    // Check if it still wants pieces from us.
	    if (!peer.isInterested() && peer.isUpLoading()
		&& coordinator.uploaders >= PeerCoordinator.MAX_UPLOADERS)
	      {
		if (Snark.debug)
		  Snark.debug("Removing uninterested uploader: " + peer);
		peer.upload(false);
		coordinator.uploaders--;
		// Put it at the back of the list
		it.remove();
		removed.add(peer);
	      }
	    else if (peer.isChoked() && peer.isDownLoading()
		     &&
		   coordinator.downloaders >= PeerCoordinator.MAX_DOWNLOADERS)
	      {
		// If they are choking us make someone else a downloader
		if (Snark.debug)
		  Snark.debug("Removing choking downloader: " + peer);
		peer.download(false);
		coordinator.downloaders--;
		// Put it at the back of the list
		it.remove();
		removed.add(peer);
	      }
	    else if (peer.isDownLoading() && download == 0
		     &&
		   coordinator.downloaders >= PeerCoordinator.MAX_DOWNLOADERS)
	      {
		// We are downloading but didn't receive anything...
		if (Snark.debug)
		  Snark.debug("Removing downloader that doesn't deliver:"
				   + peer);
		peer.download(false);
		coordinator.downloaders--;
		// Put it at the back of the list
		it.remove();
		removed.add(peer);
	      }
	    else if (peer.isUpLoading() && download < worstdownload)
	      {
		// Make sure download is good if we are uploading to the peer.
		worstdownload = download;
		worstDownloader = peer;
	      }
	  }

	// Print some statistics
	long total = coordinator.getDownloaded()/1024;
	String totalDown;
	if (total >= 10240)
	  totalDown = (total/1024) + "MB";
	else
	  totalDown = total + "KB";
	total = coordinator.getUploaded()/1024;
	String totalUp;
	if (total >= 10240)
	  totalUp = (total/1024) + "MB";
	else
	  totalUp = total + "KB";

	int needP = coordinator.storage.needed();
	long needMB
	  = needP * coordinator.metainfo.getPieceLength(0) / (1024 * 1024);
	int totalP = coordinator.metainfo.getPieces();
	long totalMB = coordinator.metainfo.getTotalLength() / (1024 * 1024);
	
	System.out.println();
	System.out.println("Down: " + downloaded/KILOPERSECOND + "KB/s"
			   + " (" + totalDown + ")"
			   + " Up: " + uploaded/KILOPERSECOND + "KB/s"
			   + " (" + totalUp + ")"
			   + " Need " + needP
			   + " (" + needMB + "MB)"
			   + " of " + totalP
			   + " (" + totalMB + "MB)"
			   + " pieces");
	System.out.println("Download #" + downloaders
			   + " Upload #" + uploaders
			   + " Interested #" + interested
			   + " Interesting #" + interesting
			   + " Choking #" + choking
			   + " Choked #" + choked);
	System.out.println();
	
	// Remove the worst downloader if needed.
	if (coordinator.uploaders >= PeerCoordinator.MAX_UPLOADERS
	    && interested > PeerCoordinator.MAX_UPLOADERS
	    && worstDownloader != null)
	  {
	    if (Snark.debug)
	      Snark.debug("Removed worst downloader: " + worstDownloader);
	    coordinator.uploaders--;
	    worstDownloader.upload(false);
	    // Put it at the back of the list
	    coordinator.peers.remove(worstDownloader);
	    removed.add(worstDownloader);
	  }
	
	// Put peers back at the end of the list that we removed earlier.
	coordinator.peers.addAll(removed);
	
	// Optimistically unchoke a peer
	coordinator.addUpDownloader();
      }
  }
}
